# Copyright 2020-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Standard two qubit gates
"""

from typing import Iterator, Union

import numpy as np

from .. import tensors, utils, var
from ..config import CONJ, CTRL, SQRT, SWAP_TARGET, TARGET
from ..gates import unitary_from_hamiltonian
from ..ops import StdGate
from ..paulialgebra import Pauli, sX, sY, sZ
from ..qubits import Qubit
from ..states import State
from ..tensors import QubitTensor
from ..utils import cached_property
from ..var import PI, Variable
from .stdgates_1q import S_H, V_H, H, I, S, V, X, Y, Z

# 2 qubit gates, alphabetic order

__all__ = (
    "B",
    "Barenco",
    "Can",
    "CH",
    "CNot",
    "CNotPow",
    "CrossResonance",
    "CV",
    "CV_H",
    "CY",
    "CYPow",
    "CZ",
    "CZPow",
    "ECP",
    "Exch",
    "Givens",
    "ISwap",
    "SqrtISwap",
    "SqrtISwap_H",
    "SqrtSwap",
    "SqrtSwap_H",
    "Swap",
    "W",
    "XX",
    "XY",
    "YY",
    "ZZ",
)


class B(StdGate):
    """The B (Berkeley) gate. Equivalent to Can(-1/2, -1/4, 0)"""

    cv_interchangeable = True

    def __init__(self, q0: Qubit, q1: Qubit):
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return Can(-1 / 2, -1 / 4, 0, *self.qubits).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        U = (
            np.asarray(
                [
                    [1 + np.sqrt(2), 0, 0, 1j],
                    [0, 1, 1j * (1 + np.sqrt(2)), 0],
                    [0, 1j * (1 + np.sqrt(2)), 1, 0],
                    [1j, 0, 0, 1 + np.sqrt(2)],
                ]
            )
            * np.sqrt(2 - np.sqrt(2))
            / 2
        )
        return tensors.asqutensor(U)

    @property
    def H(self) -> "Can":
        return self ** -1

    def __pow__(self, t: Variable) -> "Can":
        return Can(-t / 2, -t / 4, 0, *self.qubits)


class Barenco(StdGate):
    """A universal two-qubit gate:

    Ref:
        A Universal Two–Bit Gate for Quantum Computation, A. Barenco (1996)
        https://arxiv.org/pdf/quant-ph/9505016.pdf
    """

    _diagram_labels = ["───" + CTRL + "───", "Barenco({phi}, {alpha}, {theta})"]

    # Note: parameter order as defined by original paper
    def __init__(
        self,
        phi: Variable,
        alpha: Variable,
        theta: Variable,
        q0: Qubit,
        q1: Qubit,
    ) -> None:
        super().__init__(params=[phi, alpha, theta], qubits=[q0, q1])

    @cached_property
    def tensor(self) -> QubitTensor:
        phi = self.float_param("phi")
        alpha = self.float_param("alpha")
        theta = self.float_param("theta")

        unitary = [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [
                0,
                0,
                np.exp(1j * alpha) * np.cos(theta),
                -1j * np.exp(1j * (alpha - phi)) * np.sin(theta),
            ],
            [
                0,
                0,
                -1j * np.exp(1j * (alpha + phi)) * np.sin(theta),
                np.exp(1j * alpha) * np.cos(theta),
            ],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Barenco":
        phi, alpha, theta = self.params
        alpha = -alpha
        phi = phi + PI
        return Barenco(phi, alpha, theta, *self.qubits)


# TODO: Add references and explanation
# DOCME: Comment on sign conventions.
class Can(StdGate):
    r"""The canonical 2-qubit gate

    The canonical decomposition of 2-qubits gates removes local 1-qubit
    rotations, and leaves only the non-local interactions.

    .. math::
        \text{CAN}(t_x, t_y, t_z) \equiv
            \exp\Big\{-i\frac{\pi}{2}(t_x X\otimes X
            + t_y Y\otimes Y + t_z Z\otimes Z)\Big\}

    """
    cv_interchangeable = True

    def __init__(
        self, tx: Variable, ty: Variable, tz: Variable, q0: Qubit, q1: Qubit
    ) -> None:
        super().__init__(params=[tx, ty, tz], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        tx, ty, tz = self.params
        q0, q1 = self.qubits
        return (
            (tx * sX(q0) * sX(q1) + ty * sY(q0) * sY(q1) + tz * sZ(q0) * sZ(q1))
            * PI
            / 2
        )

    @cached_property
    def tensor(self) -> QubitTensor:
        tx, ty, tz = [var.asfloat(v) for v in self.params]
        xx = XX(tx, 0, 1)
        yy = YY(ty, 0, 1)
        zz = ZZ(tz, 0, 1)

        gate = yy @ xx
        gate = zz @ gate
        return gate.tensor

    @property
    def H(self) -> "Can":
        return self ** -1

    def __pow__(self, t: Variable) -> "Can":
        tx, ty, tz = self.params
        return Can(tx * t, ty * t, tz * t, *self.qubits)

    # TODO: XY, ECP, ...
    def specialize(self) -> StdGate:
        qbs = self.qubits
        tx, ty, tz = [t % 2 for t in self.params]

        tx_zero = var.isclose(tx, 0.0) or var.isclose(tx, 2.0)
        ty_zero = var.isclose(ty, 0.0) or var.isclose(ty, 2.0)
        tz_zero = var.isclose(tz, 0.0) or var.isclose(tz, 2.0)

        if ty_zero and tz_zero:
            return XX(tx, *qbs).specialize()
        elif tx_zero and tz_zero:
            return YY(ty, *qbs).specialize()
        elif tx_zero and ty_zero:
            return ZZ(tz, *qbs).specialize()
        elif np.isclose(tx, ty) and np.isclose(tx, tz):
            return Exch(tx, *qbs).specialize()
        return self


# end class Can


class CH(StdGate):
    r"""A controlled-Hadamard gate

    Equivalent to ``controlled_gate(H())`` and locally equivalent to
    ``Can(1/2, 0, 0)``

    .. math::
        \text{CH}() =
            \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \tfrac{1}{\sqrt{2}} &  \tfrac{1}{\sqrt{2}} \\
                0 & 0 & \tfrac{1}{\sqrt{2}} & -\tfrac{1}{\sqrt{2}}
            \end{pmatrix}
    """
    _diagram_labels = [CTRL, "H"]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        return H(q1).hamiltonian * (1 - sZ(q0)) / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        unitary = np.asarray(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1 / np.sqrt(2), 1 / np.sqrt(2)],
                [0, 0, 1 / np.sqrt(2), -1 / np.sqrt(2)],
            ]
        )
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CH":
        return self  # Hermitian

    # TODO: __pow__


# end class CH


class CNot(StdGate):
    r"""A controlled-not gate

    Equivalent to ``controlled_gate(X())``, and
    locally equivalent to ``Can(1/2, 0, 0)``

     .. math::
        \text{CNot}() \equiv \begin{pmatrix} 1&0&0&0 \\ 0&1&0&0 \\
                                            0&0&0&1 \\ 0&0&1&0 \end{pmatrix}
    """
    cv_tensor_structure = "permutation"
    _diagram_labels = [CTRL, TARGET]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        return X(q1).hamiltonian * (1 - sZ(q0)) / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        unitary = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CNot":
        return self  # Hermitian

    def __pow__(self, t: Variable) -> "CNotPow":
        return CNotPow(t, *self.qubits)

    def run(self, ket: State) -> State:
        axes = ket.qubit_indices(self.qubits)
        s10 = utils.multi_slice(axes, [1, 0])
        s11 = utils.multi_slice(axes, [1, 1])
        tensor = ket.tensor.copy()
        tensor[s10] = ket.tensor[s11]
        tensor[s11] = ket.tensor[s10]
        return State(tensor, ket.qubits, ket.memory)


# end class CNot


# FIXME: matrix in docs looks wrong?
class CNotPow(StdGate):
    r"""Powers of the CNot gate.

    Equivalent to ``controlled_gate(TX(t))``, and locally equivalent to
    ``Can(t/2, 0 ,0)``.

    .. math::
        \text{CTX}(t) =
            \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \cos(\frac{\theta}{2}) e^{i\frac{\theta}{2}}
                      & -i \sin(\frac{\theta}{2}) e^{i\frac{\theta}{2}} \\
                0 & 0 & -i \sin(\frac{\theta}{2}) e^{i\frac{\theta}{2}}
                      & \cos(\frac{\theta}{2}) e^{i\frac{\theta}{2}}
            \end{pmatrix}

    args:
        t:  turns (powers of the CNot gate, or controlled-powers of X)
        q0: control qubit
        q1: target qubit
    """

    _diagram_labels = [CTRL, "X^{t}"]

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        return CNot(*self.qubits).hamiltonian * t

    @cached_property
    def tensor(self) -> QubitTensor:
        ctheta = np.pi * var.asfloat(self.param("t"))
        phase = np.exp(0.5j * ctheta)
        cht = np.cos(ctheta / 2)
        sht = np.sin(ctheta / 2)
        unitary = [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, phase * cht, phase * -1.0j * sht],
            [0, 0, phase * -1.0j * sht, phase * cht],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CNotPow":
        return self ** -1

    def __pow__(self, e: Variable) -> "CNotPow":
        (t,) = self.params
        return CNotPow(e * t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (t,) = self.params
        t %= 2
        if np.isclose(t, 0.0) or np.isclose(t, 2.0):
            return I(qbs[0])
        elif np.isclose(t, 1.0):
            return CNot(*qbs)
        return self


# end class CNotPow


class CrossResonance(StdGate):
    # DOCME
    # TESTME

    _diagram_labels = ["CR({s}, {b}, {c})_q0", "CR({s}, {b}, {c})_q1"]

    def __init__(self, s: Variable, b: Variable, c: Variable, q0: Qubit, q1: Qubit):
        super().__init__(qubits=(q0, q1), params=[s, b, c])

    @property
    def hamiltonian(self) -> Pauli:
        s, b, c = self.params
        q0, q1 = self.qubits
        return s * (sX(q0) - b * sZ(q0) * sX(q1) + c * sX(q1)) * PI / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        U = unitary_from_hamiltonian(self.hamiltonian, self.qubits)
        return tensors.asqutensor(U.tensor)

    @property
    def H(self) -> "CrossResonance":
        return self ** -1

    # TESTME
    def __pow__(self, t: Variable) -> "CrossResonance":
        s, b, c = self.params
        return CrossResonance(t * s, b, c, *self.qubits)


# end class CrossResonance


class CY(StdGate):
    r"""A controlled-Y gate

    Equivalent to ``controlled_gate(Y())`` and locally equivalent to
    ``Can(1/2,0,0)``

    .. math::
        \text{CY}() =
            \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & -i \\
                0 & 0 & i & 0
            \end{pmatrix}
    """
    cv_tensor_structure = "monomial"
    _diagram_labels = [CTRL, "Y"]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        return Y(q1).hamiltonian * (1 - sZ(q0)) / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        unitary = np.asarray(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1j], [0, 0, 1j, 0]]
        )
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CY":
        return self  # Hermitian

    def __pow__(self, e: Variable) -> "CYPow":
        return CYPow(e, *self.qubits)


# end class CY


class CYPow(StdGate):
    r"""Powers of the controlled-Y gate

    Locally equivalent to ``Can(t/2, 0, 0)``

    """
    _diagram_labels = [CTRL, "Y^{t}"]

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        return CY(*self.qubits).hamiltonian * t

    @cached_property
    def tensor(self) -> QubitTensor:
        ctheta = np.pi * var.asfloat(self.param("t"))
        phase = np.exp(0.5j * ctheta)
        unitary = [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, phase * np.cos(ctheta / 2.0), phase * -np.sin(ctheta / 2.0)],
            [0, 0, phase * np.sin(ctheta / 2.0), phase * np.cos(ctheta / 2.0)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CYPow":
        return self ** -1

    def __pow__(self, e: Variable) -> "CYPow":
        (t,) = self.params
        return CYPow(e * t, *self.qubits)

    def decompose(self) -> Iterator[Union[S, S_H, CNotPow]]:
        """Convert powers of CZ gate to powers of CNOT gate"""
        (t,) = self.params
        q0, q1 = self.qubits
        yield S_H(q1)
        yield CNot(q0, q1) ** t
        yield S(q1)


# End class CYPow


class CV(StdGate):
    r"""A controlled V (sqrt of CNOT) gate."""

    _diagram_labels = [CTRL, "V"]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return CNot(*self.qubits).hamiltonian / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        q0, q1 = self.qubits
        from ..modules import ControlGate

        return ControlGate([q0], V(q1)).tensor

    @property
    def H(self) -> "CV_H":
        return CV_H(*self.qubits)

    def __pow__(self, t: Variable) -> "CNotPow":
        return CNotPow(t / 2, *self.qubits)


# end class CV


class CV_H(StdGate):
    r"""A controlled V (sqrt of CNOT) gate."""

    _diagram_labels = [CTRL, "V" + CONJ]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return -CNot(*self.qubits).hamiltonian / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        q0, q1 = self.qubits
        from ..modules import ControlGate

        return ControlGate([q0], V_H(q1)).tensor

    @property
    def H(self) -> "CV":
        return CV(*self.qubits)

    def __pow__(self, t: Variable) -> "CNotPow":
        return CNotPow(-t / 2, *self.qubits)


# end class CV_H


class CZ(StdGate):
    r"""A controlled-Z gate

    Equivalent to ``controlled_gate(Z())`` and locally equivalent to
    ``Can(1/2, 0, 0)``

    .. math::
        \text{CZ}() = \begin{pmatrix} 1&0&0&0 \\ 0&1&0&0 \\
                                    0&0&1&0 \\ 0&0&0&-1 \end{pmatrix}
    """
    cv_interchangeable = True
    cv_tensor_structure = "diagonal"
    _diagram_labels = [CTRL, CTRL]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        return Z(q1).hamiltonian * (1 - sZ(q0)) / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        unitary = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CZ":
        return self  # Hermitian

    def __pow__(self, t: Variable) -> "CZPow":
        return CZPow(t, *self.qubits)

    def run(self, ket: State) -> State:
        axes = ket.qubit_indices(self.qubits)
        s11 = utils.multi_slice(axes, [1, 1])
        tensor = ket.tensor.copy()
        tensor[s11] *= -1
        return State(tensor, ket.qubits, ket.memory)


# End class CZ


class CZPow(StdGate):
    r"""Powers of the controlled-Z gate

    Locally equivalent to ``Can(t/2, 0, 0)``

    .. math::
        \text{CZ}^t = \begin{pmatrix}
                        1 & 0 & 0 & 0 \\
                        0 & 1 & 0 & 0 \\
                        0 & 0 & 1 & 0 \\
                        0 & 0 & 0 & \exp(i \pi t)
                      \end{pmatrix}
    """
    cv_interchangeable = True
    cv_tensor_structure = "diagonal"
    _diagram_labels = [CTRL, "Z^{t}"]

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        return CZ(*self.qubits).hamiltonian * t

    @cached_property
    def tensor(self) -> QubitTensor:
        t = var.asfloat(self.param("t"))
        unitary = [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, np.exp(1j * t * np.pi)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "CZPow":
        return self ** -1

    def __pow__(self, e: Variable) -> "CZPow":
        (t,) = self.params
        return CZPow(e * t, *self.qubits)


# End class CZPow


# TESTME
# DOCME
# TODO: Add citations x2
class ECP(StdGate):
    r"""The ECP gate. The peak of the pyramid of gates in the Weyl chamber
    that can be created with a square-root of iSWAP sandwich.

    Equivalent to ``Can(1/2, 1/4, 1/4)``.
    """
    cv_interchangeable = True

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return Can(1 / 2, 1 / 4, 1 / 4, *self.qubits).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        return Can(1 / 2, 1 / 4, 1 / 4, *self.qubits).tensor

    @property
    def H(self) -> "Can":
        return self ** -1

    def __pow__(self, t: Variable) -> "Can":
        return Can(t / 2, t / 4, t / 4, *self.qubits)


# end class ECP


class Exch(StdGate):
    r"""A 2-qubit parametric gate generated from an exchange interaction.

    Equivalent to Can(t,t,t)

    """
    cv_interchangeable = True

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        return Can(t, t, t, *self.qubits).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        (t,) = self.params
        return Can(t, t, t, *self.qubits).tensor

    @property
    def H(self) -> "Exch":
        return self ** -1

    def __pow__(self, e: Variable) -> "Exch":
        (t,) = self.params
        return Exch(e * t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (t,) = self.params
        t %= 2
        if np.isclose(t, 0.0) or np.isclose(t, 2.0):
            return I(qbs[0])
        return self


# end class Exch


class Givens(StdGate):
    r"""
    In quantum computational chemistry refers to a 2-qubit gate that defined as

        Givens(θ) ≡ exp(-i θ (Y⊗X - X⊗Y) / 2)

    Locally equivalent to the XY gate.

    ::
        ───Givens(θ)_0───     ───T⁺───XY^θ/π───T────
              │                        │
        ───Givens(θ)_1───     ───T────XY^θ/π───T⁺───

    .. math::
        \text{Givens}(\theta) =
            \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & \cos(\theta) & -\sin(\theta) & 0 \\
                0 & -sin(\theta)  & \cos(\theta) & 0 \\
                0 & 0 & 0 & 1)}
            \end{pmatrix}

    """
    # Kudos: Adapted from cirq

    def __init__(self, theta: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[theta], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (theta,) = self.params
        q0, q1 = self.qubits
        return theta * (sY(q0) * sX(q1) - sX(q0) * sY(q1)) / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        unitary = [
            [1, 0, 0, 0],
            [0, np.cos(theta), -np.sin(theta), 0],
            [0, np.sin(theta), np.cos(theta), 0],
            [0, 0, 0, 1],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Givens":
        return self ** -1

    def __pow__(self, t: Variable) -> "Givens":
        (theta,) = self.params
        return Givens(t * theta, *self.qubits)


# end class Givens


class ISwap(StdGate):
    r"""A 2-qubit iSwap gate

    Equivalent to ``Can(-1/2,-1/2,0)``.

    .. math::
        \text{ISWAP}() \equiv
        \begin{pmatrix} 1&0&0&0 \\ 0&0&i&0 \\ 0&i&0&0 \\ 0&0&0&1 \end{pmatrix}

    """
    cv_interchangeable = True
    cv_tensor_structure = "monomial"
    _diagram_labels = ["iSwap"]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return Can(-1 / 2, -1 / 2, 0, *self.qubits).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        unitary = [[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]]

        return tensors.asqutensor(unitary)

    def run(self, ket: State) -> State:
        tensor = ket.tensor.copy()
        axes = ket.qubit_indices(self.qubits)
        s10 = utils.multi_slice(axes, [1, 0])
        s01 = utils.multi_slice(axes, [0, 1])
        tensor[s01] = 1.0j * ket.tensor[s10]
        tensor[s10] = 1.0j * ket.tensor[s01]
        return State(tensor, ket.qubits, ket.memory)

    @property
    def H(self) -> "XY":
        return self ** -1

    def __pow__(self, t: Variable) -> "XY":
        return XY(-t / 2, *self.qubits)


# end class ISWAP


# TESTME
class SqrtISwap(StdGate):
    r"""A square root of the iswap gate

    Equivalent to ``Can(-1/4,-1/4,0)``.
    """
    cv_interchangeable = True
    _diagram_labels = [SQRT + "iSwap"]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return Can(-1 / 4, -1 / 4, 0, *self.qubits).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        return Can(-1 / 4, -1 / 4, 0, *self.qubits).tensor

    @property
    def H(self) -> "SqrtISwap_H":
        return SqrtISwap_H(*self.qubits)

    def __pow__(self, t: Variable) -> "XY":
        return XY(-t / 4, *self.qubits)


# end class SqrtISwap


# TESTME
class SqrtISwap_H(StdGate):
    r"""The Hermitian conjugate of the square root iswap gate

    Equivalent to ``Can(1/4, 1/4, 0)``.
    """
    cv_interchangeable = True
    _diagram_labels = [SQRT + "iSwap" + CONJ]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return Can(1 / 4, 1 / 4, 0, *self.qubits).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        return Can(1 / 4, 1 / 4, 0, *self.qubits).tensor

    @property
    def H(self) -> "SqrtISwap":
        return SqrtISwap(*self.qubits)

    def __pow__(self, t: Variable) -> "XY":
        return XY(t / 4, *self.qubits)


# end class SqrtISwap_H


# TESTME
class SqrtSwap(StdGate):
    r"""Square root of the 2-qubit swap gate

    Equivalent to ``Can(1/4, 1/4, 1/4)``.
    """
    cv_interchangeable = True
    _diagram_labels = [SQRT + "Swap"]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return Can(1 / 4, 1 / 4, 1 / 4, *self.qubits).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        return Can(1 / 4, 1 / 4, 1 / 4, *self.qubits).tensor

    @property
    def H(self) -> "SqrtSwap_H":
        return SqrtSwap_H(*self.qubits)

    def __pow__(self, t: Variable) -> "Exch":
        return Exch(t / 4, *self.qubits)


# end class SqrtSwap


# TESTME
class SqrtSwap_H(StdGate):
    r"""The conjugate of the Square root swap gate

    Equivalent to ``Can(-1/4, -1/4, -1/4)``, and locally equivalent to
    ``Can(3/4, 1/4, 1/4)``
    """
    cv_interchangeable = True
    _diagram_labels = [SQRT + "Swap" + CONJ]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        return Can(-1 / 4, -1 / 4, -1 / 4, *self.qubits).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        return Can(-1 / 4, -1 / 4, -1 / 4, *self.qubits).tensor

    @property
    def H(self) -> "SqrtSwap":
        return SqrtSwap(*self.qubits)

    def __pow__(self, t: Variable) -> "Exch":
        return Exch(-t / 4, *self.qubits)


# end class SqrtSwap_H


class Swap(StdGate):
    r"""A 2-qubit swap gate

    Equivalent to ``Can(1/2, 1/2, 1/2)``.

    .. math::
        \text{SWAP}() \equiv
            \begin{pmatrix}
            1&0&0&0 \\ 0&0&1&0 \\ 0&1&0&0 \\ 0&0&0&1
            \end{pmatrix}

    """
    cv_interchangeable = True
    cv_tensor_structure = "permutation"
    _diagram_labels = [SWAP_TARGET, SWAP_TARGET]

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        return (sX(q0) * sX(q1) + sY(q0) * sY(q1) + sZ(q0) * sZ(q1) - 1) * PI / 4

    @cached_property
    def tensor(self) -> QubitTensor:
        unitary = [[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Swap":
        return self  # Hermitian

    # TESTME
    def __pow__(self, t: Variable) -> "Exch":
        return Exch(t / 2, *self.qubits)

    def run(self, ket: State) -> State:
        idx0, idx1 = ket.qubit_indices(self.qubits)
        perm = list(range(ket.qubit_nb))
        perm[idx0] = idx1
        perm[idx1] = idx0
        tensor = np.transpose(ket.tensor, perm)
        return State(tensor, ket.qubits, ket.memory)


# end class SWAP


# TESTME DOCME
class W(StdGate):
    r"""A dual-rail Hadamard gate.

    Locally equivalent to ECP, `Can(1/2, 1/4, 1/4)`.

   .. math::
        \text{W} \equiv \begin{pmatrix} 1&0&0&0 \\
        0&\tfrac{1}{\sqrt{2}}&\tfrac{1}{\sqrt{2}}&0 \\
        0&\tfrac{1}{\sqrt{2}}&-\tfrac{1}{\sqrt{2}}&0 \\
        0&0&0&1 \end{pmatrix}

    ::
        ───X───●───●───●───X───
           │   │   │   │   │
        ───●───X───H───X───●───

    Notably, the W gate diagonalizes the swap gate.
    ::
        ──x──   ───W_q0───●───Z───W_q0───
          │   =    │      │       │
        ──x──   ───W_q1───●───────W_q1───

    Refs:
        https://arxiv.org/pdf/quant-ph/0209131.pdf
        https://arxiv.org/pdf/1206.0758v3.pdf
        https://arxiv.org/pdf/1505.06552.pdf

    """
    # Kudos: Adapted from Quipper

    def __init__(self, q0: Qubit, q1: Qubit) -> None:
        super().__init__(qubits=[q0, q1])

    @cached_property
    def tensor(self) -> QubitTensor:
        rs2 = 1 / np.sqrt(2)
        unitary = [[1, 0, 0, 0], [0, rs2, rs2, 0], [0, rs2, -rs2, 0], [0, 0, 0, 1]]

        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "W":
        return self


# end class W


class XX(StdGate):
    r"""A parametric 2-qubit gate generated from an XX interaction,

    Equivalent to ``Can(t, 0, 0)``.

    XX(1/2) is the Mølmer-Sørensen gate.

    Ref: Sørensen, A. & Mølmer, K. Quantum computation with ions in thermal
    motion. Phys. Rev. Lett. 82, 1971–1974 (1999)

    Args:
        t:
    """
    # TODO: Is XX(1/2) MS gate, or is it XX(-1/2)???
    cv_interchangeable = True
    _cv_sdiagram_labels = ["XX^{t}"]

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        return Can(t, 0, 0, *self.qubits).hamiltonian

    # FIXME: Phase
    @cached_property
    def tensor(self) -> QubitTensor:
        theta = np.pi * var.asfloat(self.param("t"))
        unitary = [
            [np.cos(theta / 2), 0, 0, -1.0j * np.sin(theta / 2)],
            [0, np.cos(theta / 2), -1.0j * np.sin(theta / 2), 0],
            [0, -1.0j * np.sin(theta / 2), np.cos(theta / 2), 0],
            [-1.0j * np.sin(theta / 2), 0, 0, np.cos(theta / 2)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "XX":
        return self ** -1

    def __pow__(self, e: Variable) -> "XX":
        (t,) = self.params
        return XX(e * t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (t,) = self.params
        t %= 2
        if np.isclose(t, 0.0) or np.isclose(t, 2.0):
            return I(qbs[0])
        return self


class XY(StdGate):
    r"""XY interaction gate.

    Powers of the iSWAP gate. Equivalent to ``Can(t, t, 0)``.
    """
    cv_interchangeable = True
    _diagram_labels = ["XY^{t}"]

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        q0, q1 = self.qubits
        return t * (sX(q0) * sX(q1) + sY(q0) * sY(q1)) * PI / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        t = var.asfloat(self.param("t"))
        return Can(t, t, 0, *self.qubits).tensor

    @property
    def H(self) -> "XY":
        return self ** -1

    def __pow__(self, e: Variable) -> "XY":
        (t,) = self.params
        return XY(e * t, *self.qubits)


class YY(StdGate):
    r"""A parametric 2-qubit gate generated from a YY interaction.

    Equivalent to ``Can(0, t, 0)``, and locally equivalent to
    ``Can(t, 0, 0)``

    Args:
        t:
    """
    cv_interchangeable = True
    _diagram_labels = ["YY^{t}"]

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        q0, q1 = self.qubits
        return t * sY(q0) * sY(q1) * PI / 2

    @cached_property
    def tensor(self) -> QubitTensor:
        theta = np.pi * var.asfloat(self.param("t"))
        unitary = [
            [np.cos(theta / 2), 0, 0, 1.0j * np.sin(theta / 2)],
            [0, np.cos(theta / 2), -1.0j * np.sin(theta / 2), 0],
            [0, -1.0j * np.sin(theta / 2), np.cos(theta / 2), 0],
            [1.0j * np.sin(theta / 2), 0, 0, np.cos(theta / 2)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "YY":
        return self ** -1

    def __pow__(self, e: Variable) -> "YY":
        (t,) = self.params
        return YY(e * t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (t,) = self.params
        t %= 2
        if np.isclose(t, 0.0) or np.isclose(t, 2.0):
            return I(qbs[0])
        return self


class ZZ(StdGate):
    r"""A parametric 2-qubit gate generated from a ZZ interaction.

    Equivalent to ``Can(0,0,t)``, and locally equivalent to
    ``Can(t,0,0)``

    Args:
        t:
    """
    cv_interchangeable = True
    cv_tensor_structure = "diagonal"
    _diagram_labels = ["ZZ^{t}"]

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        q0, q1 = self.qubits
        return t * sZ(q0) * sZ(q1) * PI / 2

    # FIXME: Phase?
    @cached_property
    def tensor(self) -> QubitTensor:
        theta = np.pi * var.asfloat(self.param("t"))
        unitary = [
            [
                [[np.exp(-1j * theta / 2), 0], [0, 0]],
                [[0, np.exp(1j * theta / 2)], [0, 0]],
            ],
            [
                [[0, 0], [np.exp(1j * theta / 2), 0]],
                [[0, 0], [0, np.exp(-1j * theta / 2)]],
            ],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "ZZ":
        return self ** -1

    def __pow__(self, e: Variable) -> "ZZ":
        (t,) = self.params
        return ZZ(e * t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (t,) = self.params
        # if variable_is_symbolic(t):
        #     return self
        t = t % 2
        if np.isclose(t, 0.0) or np.isclose(t, 2.0):
            return I(qbs[0])
        return self


# end class ZZ

# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# Command line interface for the about() function
# > python -m python_template.about
#
# NB: This module should not be imported by any other code in the package
# (else we will get multiple import warnings)

if __name__ == "__main__":
    import quantumflow as qf

    qf.about()



#!/usr/bin/env python

# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Example visualizations of Circuits using LaTeX.
Requires external dependencies `pdflatex` and `poppler`.
"""

import shutil
import sys

import quantumflow as qf

if shutil.which("pdflatex") is None:
    print("Failed: External dependency `pdflatex` not found.")
    sys.exit()

if shutil.which("pdftocairo") is None:
    print("Failed: External dependency `pdftocairo` not found. " "Install `poppler`.")
    sys.exit()

# Display Bell state preparation
qf.circuit_to_image(qf.ghz_circuit([0, 1])).show()

# 4-qubit GHZ state preparation
qf.circuit_to_image(qf.ghz_circuit([0, 1, 2, 3])).show()

# 4-qubit ripple add
circ = qf.addition_circuit(
    ["a[0]", "a[1]", "a[2]", "a[3]"], ["b[0]", "b[1]", "b[2]", "b[3]"], ["cin", "cout"]
)
order = ["cin", "a[0]", "b[0]", "a[1]", "b[1]", "a[2]", "b[2]", "a[3]", "b[3]", "cout"]
latex = qf.circuit_to_latex(circ, order)
img = qf.latex_to_image(latex)
img.show()



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


import pytest
from numpy import pi

pytest.importorskip("cirq")
import cirq as cq  # noqa: E402

import quantumflow as qf  # noqa: E402
from quantumflow.xcirq import CirqSimulator  # noqa: E402
from quantumflow.xcirq import circuit_to_cirq  # noqa: E402
from quantumflow.xcirq import cirq_to_circuit  # noqa: E402
from quantumflow.xcirq import from_cirq_qubit  # noqa: E402
from quantumflow.xcirq import to_cirq_qubit  # noqa: E402


def test_from_cirq_qubit() -> None:
    assert from_cirq_qubit(cq.GridQubit(2, 3)) == (2, 3)
    assert from_cirq_qubit(cq.NamedQubit("A Name")) == "A Name"
    assert from_cirq_qubit(cq.LineQubit(101)) == 101


def test_to_cirq_qubit() -> None:
    assert cq.GridQubit(2, 3) == to_cirq_qubit((2, 3))
    assert cq.NamedQubit("A Name") == to_cirq_qubit("A Name")
    assert cq.LineQubit(101) == to_cirq_qubit(101)


def test_cirq_to_circuit() -> None:
    q0 = cq.LineQubit(0)
    q1 = cq.LineQubit(1)
    q2 = cq.LineQubit(2)

    gate = cirq_to_circuit(cq.Circuit(cq.X(q0)))[0]
    assert isinstance(gate, qf.X)
    assert gate.qubits == (0,)

    gate = cirq_to_circuit(cq.Circuit(cq.X(q1) ** 0.4))[0]
    assert isinstance(gate, qf.XPow)
    assert gate.qubits == (1,)

    gate = cirq_to_circuit(cq.Circuit(cq.CZ(q1, q0)))[0]
    assert isinstance(gate, qf.CZ)
    assert gate.qubits == (1, 0)

    gate = cirq_to_circuit(cq.Circuit(cq.CZ(q1, q0) ** 0.3))[0]
    assert isinstance(gate, qf.CZPow)
    assert gate.qubits == (1, 0)
    assert gate.param("t") == 0.3

    gate = cirq_to_circuit(cq.Circuit(cq.CNOT(q0, q1)))[0]
    assert isinstance(gate, qf.CNot)
    assert gate.qubits == (0, 1)

    gate = cirq_to_circuit(cq.Circuit(cq.CNOT(q0, q1) ** 0.25))[0]
    assert isinstance(gate, qf.CNotPow)
    assert gate.qubits == (0, 1)
    assert gate.param("t") == 0.25

    gate = cirq_to_circuit(cq.Circuit(cq.SWAP(q0, q1)))[0]
    assert isinstance(gate, qf.Swap)

    gate = cirq_to_circuit(cq.Circuit(cq.ISWAP(q0, q1)))[0]
    assert isinstance(gate, qf.ISwap)

    gate = cirq_to_circuit(cq.Circuit(cq.CSWAP(q0, q1, q2)))[0]
    assert isinstance(gate, qf.CSwap)

    gate = cirq_to_circuit(cq.Circuit(cq.CCX(q0, q1, q2)))[0]
    assert isinstance(gate, qf.CCNot)

    gate = cirq_to_circuit(cq.Circuit(cq.CCZ(q0, q1, q2)))[0]
    assert isinstance(gate, qf.CCZ)

    gate = cirq_to_circuit(cq.Circuit(cq.I(q0)))[0]
    assert isinstance(gate, qf.I)

    gate = cirq_to_circuit(cq.Circuit(cq.XX(q0, q2)))[0]
    assert isinstance(gate, qf.XX)
    assert gate.param("t") == 1.0

    gate = cirq_to_circuit(cq.Circuit(cq.XX(q0, q2) ** 0.3))[0]
    assert isinstance(gate, qf.XX)
    assert gate.param("t") == 0.3

    gate = cirq_to_circuit(cq.Circuit(cq.YY(q0, q2)))[0]
    assert isinstance(gate, qf.YY)
    assert gate.param("t") == 1.0

    gate = cirq_to_circuit(cq.Circuit(cq.YY(q0, q2) ** 0.3))[0]
    assert isinstance(gate, qf.YY)
    assert gate.param("t") == 0.3

    gate = cirq_to_circuit(cq.Circuit(cq.ZZ(q0, q2)))[0]
    assert isinstance(gate, qf.ZZ)
    assert gate.param("t") == 1.0

    gate = cirq_to_circuit(cq.Circuit(cq.ZZ(q0, q2) ** 0.3))[0]
    assert isinstance(gate, qf.ZZ)
    assert gate.param("t") == 0.3

    # Check that cirq's parity gates are the same as QF's XX, YY, ZZ
    # up to parity
    U = (cq.XX(q0, q2) ** 0.8)._unitary_()
    gate0 = qf.Unitary(U, [0, 1])
    assert qf.gates_close(gate0, qf.XX(0.8, 0, 1))

    U = (cq.YY(q0, q2) ** 0.3)._unitary_()
    gate0 = qf.Unitary(U, [0, 1])
    assert qf.gates_close(gate0, qf.YY(0.3, 0, 1))

    U = (cq.ZZ(q0, q2) ** 0.2)._unitary_()
    gate0 = qf.Unitary(U, [0, 1])
    assert qf.gates_close(gate0, qf.ZZ(0.2, 0, 1))


def version_to_versioninfo(version):  # type: ignore
    return tuple(int(c) if c.isdigit() else c for c in version.split("."))


# Note: New interface in 0.7.0
def test_cirq_to_circuit_0_7() -> None:
    q0 = cq.LineQubit(0)
    q1 = cq.LineQubit(1)
    gate = cirq_to_circuit(cq.Circuit(cq.rx(0.5).on(q0)))[0]
    assert isinstance(gate, qf.XPow)
    assert gate.param("t") == 0.5 / pi

    gate = cirq_to_circuit(cq.Circuit(cq.ry(0.5).on(q0)))[0]
    assert isinstance(gate, qf.YPow)
    assert gate.param("t") == 0.5 / pi

    gate = cirq_to_circuit(cq.Circuit(cq.rz(0.5).on(q0)))[0]
    assert isinstance(gate, qf.ZPow)
    assert gate.param("t") == 0.5 / pi

    # gate = cirq_to_circuit(cq.Circuit(cq.IdentityGate(2).on(q0, q1)))[0]

    op = (cq.PhasedISwapPowGate() ** 0.5).on(q0, q1)
    circ = cirq_to_circuit(cq.Circuit(op))
    assert qf.gates_close(circ.asgate(), qf.Givens(0.5 * pi / 2, 0, 1))

    op = cq.PhasedXZGate(
        x_exponent=0.125, z_exponent=0.25, axis_phase_exponent=0.375
    ).on(q0)
    circ = cirq_to_circuit(cq.Circuit(op))
    assert len(circ) == 3


def test_cirq_to_circuit2() -> None:

    q0 = cq.GridQubit(0, 0)
    q1 = cq.GridQubit(1, 0)

    def basic_circuit(meas=False):  # type: ignore
        sqrt_x = cq.X ** 0.5
        yield cq.X(q0) ** 0.5, sqrt_x(q1)
        yield cq.CZ(q0, q1)
        yield sqrt_x(q0), sqrt_x(q1)
        if meas:
            yield cq.measure(q0, key="q0"), cq.measure(q1, key="q1")

    cqc = cq.Circuit()
    cqc.append(basic_circuit())

    print()
    print(cqc)

    circ = cirq_to_circuit(cqc)
    print()
    print(qf.circuit_to_diagram(circ))


def test_circuit_to_circ() -> None:
    q0, q1, q2 = "q0", "q1", "q2"

    circ0 = qf.Circuit()
    circ0 += qf.I(q0)
    circ0 += qf.X(q1)
    circ0 += qf.Y(q2)

    circ0 += qf.Z(q0)
    circ0 += qf.S(q1)
    circ0 += qf.T(q2)

    circ0 += qf.H(q0)
    circ0 += qf.H(q1)
    circ0 += qf.H(q2)

    circ0 += qf.XPow(0.6, q0)
    circ0 += qf.YPow(0.6, q1)
    circ0 += qf.ZPow(0.6, q2)

    circ0 += qf.XX(0.2, q0, q1)
    circ0 += qf.YY(0.3, q1, q2)
    circ0 += qf.ZZ(0.4, q2, q0)

    circ0 += qf.CZ(q0, q1)
    circ0 += qf.CNot(q0, q1)
    circ0 += qf.Swap(q0, q1)
    circ0 += qf.ISwap(q0, q1)

    circ0 += qf.CCZ(q0, q1, q2)
    circ0 += qf.CCNot(q0, q1, q2)
    circ0 += qf.CSwap(q0, q1, q2)

    circ0 += qf.FSim(1, 2, q0, q1)

    diag0 = qf.circuit_to_diagram(circ0)
    # print()
    # print(diag0)

    cqc = circuit_to_cirq(circ0)
    # print(cqc)
    circ1 = cirq_to_circuit(cqc)

    diag1 = qf.circuit_to_diagram(circ1)
    # print()
    # print(diag1)

    assert diag0 == diag1


def test_circuit_to_circ_exception() -> None:
    circ0 = qf.Circuit([qf.Can(0.2, 0.3, 0.1, 0, 1)])
    with pytest.raises(NotImplementedError):
        circuit_to_cirq(circ0)


def test_circuit_to_circ_translate() -> None:
    circ0 = qf.Circuit([qf.Can(0.2, 0.3, 0.1, 0, 1)])
    cqc = circuit_to_cirq(circ0, translate=True)
    circ2 = cirq_to_circuit(cqc)
    assert qf.circuits_close(circ0, circ2)


def test_cirq_simulator() -> None:
    q0, q1, q2 = "q0", "q1", "q2"

    circ0 = qf.Circuit()
    circ0 += qf.I(q0)
    circ0 += qf.I(q1)
    circ0 += qf.I(q2)
    circ0 += qf.X(q1)
    circ0 += qf.Y(q2)

    circ0 += qf.Z(q0)
    circ0 += qf.S(q1)
    circ0 += qf.T(q2)

    circ0 += qf.H(q0)
    circ0 += qf.H(q1)
    circ0 += qf.H(q2)

    circ0 += qf.XPow(0.6, q0)
    circ0 += qf.YPow(0.6, q1)
    circ0 += qf.ZPow(0.6, q2)

    circ0 += qf.XX(0.2, q0, q1)
    circ0 += qf.YY(0.3, q1, q2)
    circ0 += qf.ZZ(0.4, q2, q0)

    circ0 += qf.CZ(q0, q1)
    circ0 += qf.CNot(q0, q1)
    circ0 += qf.Swap(q0, q1)
    circ0 += qf.ISwap(q0, q1)

    circ0 += qf.CCZ(q0, q1, q2)
    circ0 += qf.CCNot(q0, q1, q2)
    circ0 += qf.CSwap(q0, q1, q2)

    ket0 = qf.random_state([q0, q1, q2])
    ket1 = circ0.run(ket0)
    sim = CirqSimulator(circ0)
    ket2 = sim.run(ket0)

    assert ket1.qubits == ket2.qubits

    print(qf.state_angle(ket1, ket2))
    assert qf.states_close(ket1, ket2)

    assert qf.states_close(circ0.run(), sim.run())


def test_circuit_to_cirq_unitary() -> None:
    gate0 = qf.RandomGate([4, 5])
    circ0 = qf.Circuit([gate0])
    cqc = circuit_to_cirq(circ0)
    circ1 = cirq_to_circuit(cqc)
    assert qf.circuits_close(circ0, circ1)


# fin



# Copyright 2019-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

# DOCME

from numbers import Real
from typing import Any, Mapping, Union

import numpy as np
import sympy

from . import utils
from .config import ATOL, RTOL

__all__ = ("Variable", "PI")


Symbol = sympy.Symbol
"""Class for symbols in symbolic expressions"""


Variable = Union[Real, sympy.Expr]
"""Type for parameters. Either a float, sympy.Symbol or sympy.Expr"""


PI = sympy.pi
"""Symbolic constant pi"""


def is_symbolic(x: Any) -> bool:
    """Returns true if a symbolic expression"""
    return isinstance(x, sympy.Expr)


def isclose(x: Variable, y: Variable, atol: float = ATOL, rtol: float = RTOL) -> bool:
    """Compares two variables.

    Returns: True if variables are almost identical concrete numbers,
        of if they are the same symbolic expression, else False.
    """
    if not is_symbolic(x) and not is_symbolic(y):
        return np.isclose(x, y, atol=atol, rtol=rtol)
    if is_symbolic(x) and is_symbolic(y):
        return x == y
    return False


# DOCME # Testme
def almost_zero(x: Variable, atol: float = ATOL) -> bool:
    """Is the variable symbolically zero, or numerically almost zero."""
    if x == sympy.S.Zero:
        return True
    return isclose(x, 0.0, atol=atol)


def asfloat(x: Variable, subs: Mapping[str, float] = None) -> float:
    """Convert a variable to a float"""
    if is_symbolic(x) and subs:
        x = x.evalf(subs=subs)  # type: ignore
    return float(x)


def asexpression(flt: float) -> sympy.Expr:
    """Attempt to convert a real number into a simpler symbolic
    representation.

    Returns:
        A sympy Symbol. (Convert to string with str(sym) or to latex with
            sympy.latex(sym)
    Raises:
        ValueError:     If cannot simplify float
    """
    try:
        ratio = utils.rationalize(flt)
        res = sympy.simplify(ratio)
    except ValueError:
        try:
            ratio = utils.rationalize(flt / np.pi)
            res = sympy.simplify(ratio) * sympy.pi
        except ValueError:
            ratio = utils.rationalize(flt * np.pi)
            res = sympy.simplify(ratio) / sympy.pi
    return res


# The following math functions act on Variables, and return symbolic expression if
# the variable is symbolic, or numerical values if the variable is a number


def arccos(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.acos(x)
    return np.arccos(x)


def arcsin(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.asin(x)
    return np.arcsin(x)


def arctan(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.atan(x)
    return np.arctan(x)


def arctan2(x1: Variable, x2: Variable) -> Variable:
    if is_symbolic(x1) or is_symbolic(x2):
        return sympy.atan2(x1, x2)
    return np.arctan2(x1, x2)


def cos(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.cos(x)
    return np.cos(x)


def exp(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.exp(x)
    return np.exp(x)


def sign(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.sign(x)
    return np.sign(x)


def sin(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.sin(x)
    return np.sin(x)


def sqrt(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.sqrt(x)
    return np.sqrt(x)


def tan(x: Variable) -> Variable:
    if is_symbolic(x):
        return sympy.tan(x)
    return np.tan(x)


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""
.. contents:: :local:
.. currentmodule:: quantumflow.xcirq

Interface between Google's Cirq and QuantumFlow


.. autoclass:: CirqSimulator

.. autofunction:: cirq_to_circuit
.. autofunction:: circuit_to_cirq

.. autofunction:: from_cirq_qubit
.. autofunction:: to_cirq_qubit

"""

# Conventions
# cqc: Abbreviation for Cirq circuit

from typing import List, Type, cast

import cirq
import numpy as np

from . import var
from .circuits import Circuit
from .ops import Gate, Operation, Unitary
from .qubits import Qubit, Qubits
from .states import State, zero_state
from .stdgates import (
    CCZ,
    CZ,
    XX,
    YY,
    ZZ,
    CCNot,
    CNot,
    CSwap,
    FSim,
    H,
    I,
    ISwap,
    S,
    Swap,
    T,
    X,
    XPow,
    Y,
    YPow,
    Z,
    ZPow,
)
from .translate import circuit_translate, select_translators

__all__ = (
    "from_cirq_qubit",
    "to_cirq_qubit",
    "cirq_to_circuit",
    "circuit_to_cirq",
    "CirqSimulator",
)


CIRQ_GATES: List[Type[Gate]] = [
    I,
    X,
    Y,
    Z,
    S,
    T,
    H,
    XPow,
    YPow,
    ZPow,
    CZ,
    Swap,
    ISwap,
    CNot,
    XX,
    YY,
    ZZ,
    CCNot,
    CSwap,
    CCZ,
    FSim,
]
"""List of QuantumFlow gates that we know how to convert to Cirq"""
# TODO: Perhaps should be string names, rather than classes.


class CirqSimulator(Operation):
    """Interface to the Cirq quantum simulator. Adapts a QF Circuit (or
    other sequence of Operations). Can itself be included in Circuits,
    like any other Operation.

    Note that Cirq uses 64 bit complex floats (QF uses 128 bits), so
    results will not be as accurate.
    """

    def __init__(self, *elements: Operation) -> None:
        self._circuit = Circuit(*elements)
        self._cirq = circuit_to_cirq(self._circuit)

        # TODO: Translate gates

    @property
    def qubits(self) -> Qubits:
        return self._circuit.qubits

    def run(self, ket: State = None) -> State:
        if ket is None:
            qubits = self.qubits
            ket = zero_state(qubits=qubits)

        tensor = ket.tensor.flatten()
        tensor = np.asarray(tensor, dtype=np.complex64)
        sim = cirq.Simulator()
        res = sim.simulate(self._cirq, initial_state=tensor)
        res = cast(cirq.WaveFunctionTrialResult, res)
        tensor = res.state_vector()
        return State(tensor, ket.qubits, ket.memory)


def from_cirq_qubit(qb: cirq.Qid) -> Qubit:
    """
    Convert a cirq qubit (a subtype of Qid) into regular python type.
    A ``LineQubit`` becomes an int, a ``GridQubit`` becomes a tuple of two
    integers, and ``NamedQubit`` (and anything else) becomes a string
    """
    if isinstance(qb, cirq.LineQubit):
        return qb.x
    elif isinstance(qb, cirq.GridQubit):
        return (qb.row, qb.col)
    elif isinstance(qb, cirq.NamedQubit):
        return qb.name
    return str(qb)  # pragma: no cover


def to_cirq_qubit(qubit: Qubit) -> cirq.Qid:
    """Convert qubit names (any python object) into
    cirq qubits (subtypes of Qid). Returns either
    a LineQubit (for integers), GridQubit (for tuples of row and column),
    or a NamedQubit for all other objects.
    """
    if isinstance(qubit, int):
        return cirq.LineQubit(qubit)
    elif (
        isinstance(qubit, tuple)
        and len(qubit) == 2
        and isinstance(qubit[0], int)
        and isinstance(qubit[1], int)
    ):
        return cirq.GridQubit(row=qubit[0], col=qubit[1])
    return cirq.NamedQubit(str(qubit))


def cirq_to_circuit(cqc: cirq.Circuit) -> Circuit:
    """Convert a Cirq circuit to a QuantumFlow circuit"""

    simple_gates = {
        cirq.CSwapGate: CSwap,
        # cirq.IdentityGate: I,
    }

    exponent_gates = {
        cirq.ops.pauli_gates._PauliX: X,
        cirq.ops.pauli_gates._PauliY: Y,
        cirq.ops.pauli_gates._PauliZ: Z,
        cirq.XPowGate: X,
        cirq.YPowGate: Y,
        cirq.ZPowGate: Z,
        cirq.HPowGate: H,
        cirq.CZPowGate: CZ,
        cirq.CNotPowGate: CNot,
        cirq.SwapPowGate: Swap,
        cirq.ISwapPowGate: ISwap,
        cirq.CCXPowGate: CCNot,
        cirq.CCZPowGate: CCZ,
    }

    parity_gates = {cirq.XXPowGate: XX, cirq.YYPowGate: YY, cirq.ZZPowGate: ZZ}

    decomposable_gates = {"PhasedISwapPowGate", "PhasedXZGate"}

    circ = Circuit()
    qubit_map = {q: from_cirq_qubit(q) for q in cqc.all_qubits()}
    for op in cqc.all_operations():
        gatetype = type(op.gate)

        qbs = [qubit_map[qb] for qb in op.qubits]
        t = getattr(op.gate, "exponent", 1)

        if gatetype is cirq.IdentityGate:
            for q in qbs:
                circ += I(q)
        elif gatetype in simple_gates:
            circ += simple_gates[gatetype](*qbs)  # type: ignore
        elif gatetype in exponent_gates:
            gate = exponent_gates[gatetype](*qbs)  # type: ignore
            if t != 1:
                gate **= t
            circ += gate
        elif gatetype in parity_gates:
            circ += parity_gates[gatetype](t, *qbs)  # type: ignore
        elif gatetype.__name__ in decomposable_gates:
            subcqc = cirq.Circuit(op._decompose_())  # type: ignore
            circ += cirq_to_circuit(subcqc)
        elif gatetype.__name__ == "FSimGate":
            circ += FSim(op.gate.theta, op.gate.phi, *qbs)  # type: ignore
        elif gatetype.__name__ == "MatrixGate":
            matrix = op.gate._matrix  # type: ignore
            circ += Unitary(matrix, qbs)
        else:
            raise NotImplementedError(str(op.gate))  # pragma: no cover

    circ = circ.specialize()

    return circ


def circuit_to_cirq(circ: Circuit, translate: bool = False) -> cirq.Circuit:
    """Convert a QuantumFlow circuit to a Cirq circuit."""

    if translate:
        circ = translate_to_cirq(circ)

    qubit_map = {q: to_cirq_qubit(q) for q in circ.qubits}

    cqc = cirq.Circuit()

    operations = {
        I: cirq.I,
        X: cirq.X,
        Y: cirq.Y,
        Z: cirq.Z,
        S: cirq.S,
        T: cirq.T,
        H: cirq.H,
        CNot: cirq.CNOT,
        CZ: cirq.CZ,
        Swap: cirq.SWAP,
        ISwap: cirq.ISWAP,
        CCZ: cirq.CCZ,
        CCNot: cirq.CCX,
        CSwap: cirq.CSWAP,
    }

    # TODO: HPow -> cirq.ops.HPowGate,
    turn_gates = {
        XPow: cirq.X,
        YPow: cirq.Y,
        ZPow: cirq.Z,
        XX: cirq.XX,
        YY: cirq.YY,
        ZZ: cirq.ZZ,
    }

    for op in circ:
        qbs = [qubit_map[qb] for qb in op.qubits]

        if type(op) in operations:
            cqc.append(operations[type(op)].on(*qbs))
        elif type(op) in turn_gates:
            t = op.param("t")
            cqc.append(turn_gates[type(op)](*qbs) ** t)
        elif isinstance(op, FSim):
            theta, phi = op.params
            theta = var.asfloat(theta)
            phi = var.asfloat(phi)
            gate = cirq.FSimGate(theta=theta, phi=phi).on(*qbs)
            cqc.append(gate)
        elif isinstance(op, Unitary):
            matrix = op.asoperator()
            gate = cirq.MatrixGate(matrix).on(*qbs)
            cqc.append(gate)
        else:
            # print(type(op))
            raise NotImplementedError(repr(op))

    return cqc


def translate_to_cirq(circ: Circuit) -> Circuit:
    """Convert QF gates to gates understood by cirq"""
    target_gates = CIRQ_GATES
    trans = select_translators(target_gates)
    circ = circuit_translate(circ, trans)
    return circ


# fin



#!/usr/bin/env python

# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow Examples
"""


import quantumflow as qf


def prepare_w4():
    """
    Prepare a 4-qubit W state using sqrt(iswaps) and local gates
    """
    circ = qf.Circuit()
    circ += qf.X(1)

    circ += qf.ISwap(1, 2) ** 0.5
    circ += qf.S(2)
    circ += qf.Z(2)

    circ += qf.ISwap(2, 3) ** 0.5
    circ += qf.S(3)
    circ += qf.Z(3)

    circ += qf.ISwap(0, 1) ** 0.5
    circ += qf.S(0)
    circ += qf.Z(0)

    ket = circ.run()

    return ket


if __name__ == "__main__":

    def main():
        """CLI"""
        print(prepare_w4.__doc__)
        print("states amplitudes")
        ket = prepare_w4()
        qf.print_state(ket)

    main()



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import networkx as nx
import numpy as np
import pytest

import quantumflow as qf


def test_gradients() -> None:
    # This test only checks that code runs, not that we get correct answers
    # graph = nx.grid_graph([2, 3])
    graph = nx.grid_graph([2, 1])
    layers = 2
    params = qf.graph_circuit_params(graph, layers)
    circ = qf.graph_circuit(graph, layers, params)
    circ += qf.H((1, 1))  # add a non-parameterized gate. Should be ignored

    qubits = circ.qubits
    ket0 = qf.zero_state(qubits)
    ket1 = qf.random_state(qubits)

    grads0 = qf.state_fidelity_gradients(ket0, ket1, circ)
    # print(grads0)

    _ = qf.state_angle_gradients(ket0, ket1, circ)
    # print(grads1)

    proj = qf.Projection([ket1])
    grads2 = qf.expectation_gradients(ket0, circ, hermitian=proj)
    # print(grads2)

    # Check that qf.expectation_gradients() gives same answers for
    # fidelity as f.state_fidelity_gradients()
    for g0, g1 in zip(grads0, grads2):
        assert np.isclose(g0, g1)
        print(g0, g1)


def test_gradients_func() -> None:
    graph = nx.grid_graph([2, 1])
    layers = 2
    params = qf.graph_circuit_params(graph, layers)
    circ = qf.graph_circuit(graph, layers, params)
    circ += qf.H((1, 1))  # add a non-parameterized gate. Should be ignored
    qubits = circ.qubits
    ket0 = qf.zero_state(qubits)
    ket1 = qf.random_state(qubits)

    grads1 = qf.state_angle_gradients(ket0, ket1, circ)

    proj = qf.Projection([ket1])
    grads3 = qf.expectation_gradients(
        ket0,
        circ,
        hermitian=proj,
        dfunc=lambda fid: -1 / (2 * np.sqrt((1 - fid) * fid)),
    )
    # print(grads3)

    for g0, g1 in zip(grads1, grads3):
        assert np.isclose(g0, g1)
        print(g0, g1)


def test_gradient_errors() -> None:
    circ = qf.Circuit()
    circ += qf.CPhase(0.2, 0, 1)  # Not (currently) differentiable
    qubits = circ.qubits
    ket0 = qf.zero_state(qubits)
    ket1 = qf.random_state(qubits)

    with pytest.raises(ValueError):
        qf.state_fidelity_gradients(ket0, ket1, circ)

    with pytest.raises(ValueError):
        qf.parameter_shift_circuits(circ, 0)

    with pytest.raises(ValueError):
        qf.expectation_gradients(ket0, circ, qf.IdentityGate([0, 1]))


def test_parameter_shift_circuits() -> None:
    """Checks that gradients calculated with middle out algorithm
    match gradients calculated from parameter shift rule.
    """
    graph = nx.grid_graph([2, 2])
    layers = 2
    params = qf.graph_circuit_params(graph, layers)
    circ = qf.graph_circuit(graph, layers, params)
    N = circ.qubit_nb
    qubits = circ.qubits

    ket0 = qf.zero_state(qubits)
    ket1 = qf.random_state(qubits)
    grads = qf.state_fidelity_gradients(ket0, ket1, circ)

    for n in range(N):
        r, circ0, circ1 = qf.parameter_shift_circuits(circ, n)
        fid0 = qf.state_fidelity(circ0.run(ket0), ket1)
        fid1 = qf.state_fidelity(circ1.run(ket0), ket1)
        grad = r * (fid1 - fid0)
    assert np.isclose(grad, grads[n])


# fin
